/******************************************************
 *
 * Name:         big-johns-bbq-data-warehouse-az.sql
 *     
 * Design Phase:
 *     Author:   John Miner
 *     Date:     04-17-2015
 *     Blog:     www.craftydba.com
 *
 *     Purpose:  Create a large table of customers with
 *               1 million rows of data.
 *
 ******************************************************/


--
-- 1 - Create Bbq Data Warehouse
--

-- Which database to use.
USE [master]
GO

-- Delete existing databases.
IF EXISTS (SELECT name FROM sys.databases WHERE name = N'BIG_JONS_BBQ_DW')
    DROP DATABASE [BIG_JONS_BBQ_DW]
GO

-- Add new databases.
CREATE DATABASE [BIG_JONS_BBQ_DW]
(
MAXSIZE = 20GB,
EDITION = 'STANDARD',
SERVICE_OBJECTIVE = 'S2'
)
GO   

-- Show the new database
SELECT * FROM sys.databases WHERE name = 'BIG_JONS_BBQ_DW';
GO



--
-- 2 - Create Fact schema
--

-- Which database to use.
USE [BIG_JONS_BBQ_DW]
GO

-- Delete existing schema.
IF EXISTS (SELECT * FROM sys.schemas WHERE name = N'FACT')
DROP SCHEMA [FACT]
GO

-- Add new schema.
CREATE SCHEMA [FACT] AUTHORIZATION [dbo]
GO

-- Show the new schema
SELECT * FROM sys.schemas WHERE name = 'FACT';
GO



--
-- 3 - Create Dimension schema
--

-- Which database to use.
USE [BIG_JONS_BBQ_DW]
GO

-- Delete existing schema.
IF EXISTS (SELECT * FROM sys.schemas WHERE name = N'DIM')
DROP SCHEMA [DIM]
GO

-- Add new schema.
CREATE SCHEMA [DIM] AUTHORIZATION [dbo]
GO

-- Show the new schema
SELECT * FROM sys.schemas WHERE name = 'DIM';
GO



--
-- 4 - Create stage schema
--

-- Delete existing schema.
IF EXISTS (SELECT * FROM sys.schemas WHERE name = N'STAGE')
DROP SCHEMA [STAGE]
GO

-- Add new schema.
CREATE SCHEMA [STAGE] AUTHORIZATION [dbo]
GO

-- Show the new schema
SELECT * FROM sys.schemas WHERE name = 'STAGE';
GO



--
-- 5 - Create date dim table 
--

-- Use the correct database
USE [BIG_JONS_BBQ_DW];
GO

-- Delete existing table
IF  OBJECT_ID(N'[DIM].[DIM_DATE]') > 0
    DROP TABLE [DIM].[DIM_DATE]
GO

-- Create new table
CREATE TABLE [DIM].[DIM_DATE]
(
  date_key INT IDENTITY(1,1) NOT NULL CONSTRAINT PK_DIM_DATE PRIMARY KEY,
  date_date SMALLDATETIME NOT NULL,
  date_string VARCHAR(10) NOT NULL,
  date_month VARCHAR(3) NOT NULL,
  date_day VARCHAR(3) NOT NULL,
  date_int_qtr INT NOT NULL,
  date_int_doy INT NOT NULL,
  date_int_yyyy INT NOT NULL,
  date_int_mm INT NOT NULL,
  date_int_dd INT NOT NULL
);
GO

-- Drop existing index
IF EXISTS (SELECT name FROM sys.indexes WHERE name = N'IX_DIM_DATE_STR')
DROP INDEX IX_DIM_DATE_STR ON [DIM].[DIM_DATE];
GO

-- Create new index
CREATE INDEX IX_DIM_DATE_STR
    ON [DIM].[DIM_DATE] (date_string); 
GO

-- Import data - bcp-date-info.bat 

-- Show top 100 sample records
SELECT TOP 100 *
FROM [DIM].[DIM_DATE];
GO



--
-- 6 - Create user defined functions
--

-- Make a year + quarter hash
CREATE FUNCTION [DIM].[UFN_GET_DATE_HASH] (@VAR_DATE_STR VARCHAR(10))
  RETURNS INT
BEGIN
  RETURN 
  (
      SELECT [date_int_yyyy] * 1000 + [date_int_qtr] 
	  FROM [DIM].[DIM_DATE] WHERE [date_string] = @VAR_DATE_STR
  )
END 
GO

-- Given a string, returnt the key
CREATE FUNCTION [DIM].[UFN_GET_DATE_KEY] (@VAR_DATE_STR VARCHAR(10))
  RETURNS INT
BEGIN
  RETURN 
  (
      SELECT date_key
	  FROM [DIM].[DIM_DATE] WHERE [date_string] = @VAR_DATE_STR
  )
END 
GO



--
-- 7 - Create zip code table
--

-- http://www.unitedstateszipcodes.org/zip-code-database/


-- Delete existing table
IF  EXISTS (
    SELECT * FROM sys.objects
    WHERE object_id = OBJECT_ID(N'[STAGE].[USPS_CITY_STATE_ZIP]') AND
    type in (N'U'))
DROP TABLE [STAGE].[USPS_CITY_STATE_ZIP]
GO

-- Create new table
CREATE TABLE [STAGE].[USPS_CITY_STATE_ZIP]
(
   csz_id int not null,
   csz_zip char(5) not null,
   csz_type varchar(128),
   csz_city varchar(128),
   csz_state char(2),
   csz_areacode varchar(128),
   csz_latitude decimal,
   csz_longitude decimal,
   csz_country char(2)
);
GO

-- Alter the table (primary key)
ALTER TABLE [STAGE].[USPS_CITY_STATE_ZIP]
   ADD CONSTRAINT PK_USPS_CITY_STATE_ZIP_ID PRIMARY KEY CLUSTERED (csz_id);
GO

-- Import data - bcp-city-info.bat 

-- Remove quotes from area code
UPDATE [STAGE].[USPS_CITY_STATE_ZIP]
SET [csz_areacode] = REPLACE(ISNULL([csz_areacode], ''), '"', '');
GO

-- Show top 100 sample records
SELECT TOP 100 *
FROM [BIG_JONS_BBQ_DW].[STAGE].[USPS_CITY_STATE_ZIP];
GO



--
-- 8 - Create first names table
--

-- http://baby-names.familyeducation.com/popular-names/girls/
-- http://baby-names.familyeducation.com/popular-names/boys/

-- Delete existing table
IF  EXISTS (
    SELECT * FROM sys.objects
    WHERE object_id = OBJECT_ID(N'[STAGE].[COMMON_FIRST_NAMES]') AND
    type in (N'U'))
DROP TABLE [STAGE].[COMMON_FIRST_NAMES]
GO

-- Create new table
CREATE TABLE [STAGE].[COMMON_FIRST_NAMES]
(
   cfn_id int not null,
   cfn_gender char(1),
   cfn_fname varchar(40)
);
GO

-- Alter the table (primary key)
ALTER TABLE [STAGE].[COMMON_FIRST_NAMES]
   ADD CONSTRAINT PK_COMMON_FIRST_NAMES_ID PRIMARY KEY CLUSTERED (cfn_id);
GO

-- Import data - bcp-fname-info.bat 


-- Show all records
SELECT *
FROM [BIG_JONS_BBQ_DW].[STAGE].[COMMON_FIRST_NAMES];
GO



--
-- 9 - Create last names table
--

-- http://names.mongabay.com/most_common_surnames.htm

-- Delete existing table
IF  EXISTS (
    SELECT * FROM sys.objects
    WHERE object_id = OBJECT_ID(N'[STAGE].[COMMON_LAST_NAMES]') AND
    type in (N'U'))
DROP TABLE [STAGE].[COMMON_LAST_NAMES]
GO

-- Create new table
CREATE TABLE [STAGE].[COMMON_LAST_NAMES]
(
   cln_lname varchar(40),
   cln_count bigint,
   cln_pct real,
   cln_id int not null,
);
GO

-- Alter the table (primary key)
ALTER TABLE [STAGE].[COMMON_LAST_NAMES]
   ADD CONSTRAINT PK_COMMON_LAST_NAMES_ID PRIMARY KEY CLUSTERED (cln_id);
GO

-- Import data - bcp-lname-info.bat 

-- Show all records
SELECT *
FROM [BIG_JONS_BBQ_DW].[STAGE].[COMMON_LAST_NAMES];
GO



--
-- 10 - Create street names table
--

-- http://www.livingplaces.com/streets/most-popular_street_names.html

-- Delete existing table
IF  EXISTS (
    SELECT * FROM sys.objects
    WHERE object_id = OBJECT_ID(N'[STAGE].[COMMON_STREET_NAMES]') AND
    type in (N'U'))
DROP TABLE [STAGE].[COMMON_STREET_NAMES]
GO

-- Create new table
CREATE TABLE [STAGE].[COMMON_STREET_NAMES]
(
   csn_count int,
   csn_sname varchar(40)
);
GO

-- Alter the table (identity)
ALTER TABLE [STAGE].[COMMON_STREET_NAMES]
   ADD csn_id INT NOT NULL IDENTITY (1, 1)
GO

-- Alter the table (primary key)
ALTER TABLE [STAGE].[COMMON_STREET_NAMES]
   ADD CONSTRAINT PK_COMMON_STREET_NAMES_ID PRIMARY KEY CLUSTERED (csn_id);
GO

-- Import data - bcp-street-info.bat 

-- Show all records
SELECT *
FROM [BIG_JONS_BBQ_DW].[STAGE].[COMMON_STREET_NAMES];
GO



--
-- 11 - Create pig packages
--

-- Delete existing table
IF  EXISTS (
    SELECT * FROM sys.objects
    WHERE object_id = OBJECT_ID(N'[DIM].[PIG_PACKAGES]') AND
    type in (N'U'))
DROP TABLE [DIM].[PIG_PACKAGES]
GO

-- Create new table
CREATE TABLE [DIM].[PIG_PACKAGES]
(
   pkg_id int not null identity (1, 1),
   pkg_name varchar(40)
);
GO

-- Alter the table (primary key)
ALTER TABLE [DIM].[PIG_PACKAGES]
   ADD CONSTRAINT PK_PIG_PACKAGES_ID PRIMARY KEY CLUSTERED (pkg_id);
GO

-- Add data to the table
INSERT INTO [DIM].[PIG_PACKAGES] (pkg_name)
VALUES
   ('Baby Back Ribs'),
   ('Bacon'),
   ('Ham'),
   ('Hock'),
   ('Jowl Bacon'),
   ('Lion Chops'),
   ('Neck Chop'),
   ('Picnic Shoulder'),
   ('Pork Butt'),
   ('Spare Ribs');

-- Show all records
SELECT *
FROM [BIG_JONS_BBQ_DW].[DIM].[PIG_PACKAGES];
GO


--
-- 12 - Create a sequence object
--

-- Delete existing sequence.
IF EXISTS (SELECT name FROM sys.sequences WHERE name = N'SEQ_CUSTOMERS')
    DROP SEQUENCE [FACT].[SEQ_CUSTOMERS]
GO

-- Add new sequence.
CREATE SEQUENCE [FACT].[SEQ_CUSTOMERS]
    AS INT
    START WITH 100
    INCREMENT BY 1
    MINVALUE 1
    NO MAXVALUE 
    NO CYCLE
    NO CACHE;
GO

-- Alter the sequence.
ALTER SEQUENCE [FACT].[SEQ_CUSTOMERS] RESTART WITH 1;
GO

-- Show the new sequence
SELECT * FROM sys.sequences WHERE name = 'SEQ_CUSTOMERS' ;
GO


--
-- 13 - Create customers fact table 
--

-- Delete existing table
IF  EXISTS (
    SELECT * FROM sys.objects
    WHERE object_id = OBJECT_ID(N'[FACT].[CUSTOMERS]') AND
    type in (N'U'))
DROP TABLE [FACT].[CUSTOMERS]
GO

-- Create new table
CREATE TABLE [FACT].[CUSTOMERS]
(
   cus_id int not null,
   cus_lname varchar(40),
   cus_fname varchar(40),
   cus_phone char(12),
   cus_address varchar(40),
   cus_city varchar(20),
   cus_state char(2),
   cus_zip char(5) not null,
   cus_package_key int not null,
   cus_start_date_key int not null,
   cus_end_date_key int not null,
   cus_date_str varchar(10) not null,
   cus_qtr_key as [DIM].[UFN_GET_DATE_HASH] (cus_date_str) 
);
GO


-- Alter the customers table (primary key)
ALTER TABLE [FACT].[CUSTOMERS] 
   ADD CONSTRAINT PK_CUSTOMERS_ID PRIMARY KEY CLUSTERED (cus_id);
GO

-- Alter the customers table (foreign key - package key)
ALTER TABLE [FACT].[CUSTOMERS] 
ADD CONSTRAINT FK_DIM_PIG_PACKAGES FOREIGN KEY (cus_package_key)
    REFERENCES [DIM].[PIG_PACKAGES] (pkg_id) ;

-- Alter the customers table (foreign key - start date key)
ALTER TABLE [FACT].[CUSTOMERS] 
ADD CONSTRAINT FK_DIM_START_DATE FOREIGN KEY (cus_start_date_key)
    REFERENCES [DIM].[DIM_DATE] (date_key) ;

-- Alter the customers table (foreign key - start date key)
ALTER TABLE [FACT].[CUSTOMERS] 
ADD CONSTRAINT FK_DIM_END_DATE FOREIGN KEY (cus_end_date_key)
    REFERENCES [DIM].[DIM_DATE] (date_key) ;

-- Alter the customers table (default constraint - start date key)
ALTER TABLE [FACT].[CUSTOMERS] 
   ADD CONSTRAINT DF_START_DTE DEFAULT (1) FOR [cus_start_date_key];
GO

-- Alter the customers table (default constraint - end date key)
ALTER TABLE [FACT].[CUSTOMERS] 
   ADD CONSTRAINT DF_END_DTE DEFAULT (365) FOR [cus_end_date_key];
GO



--
-- 14 - MAIN PROGRAM - ADD 1M ROWS OF DATA (< 5m runtime)
--

-- Import data - bcp-street-info.bat

/*
999999 rows copied.
Network packet size (bytes): 4096
Clock Time (ms.) Total     : 284281 Average : (3517.64 rows per sec.)
*/

-- Remove the data

/*
TRUNCATE TABLE [FACT].[CUSTOMERS];
ALTER SEQUENCE [FACT].[SEQ_CUSTOMERS] RESTART WITH 1;
GO
*/


--
-- 16 - Show the distribution of the data
--

-- DBA's are still needed

-- Very bad performance in cloud
SELECT 
  c.cus_qtr_key, 
  count(*) as cus_total 
FROM 
  [FACT].[CUSTOMERS] c
GROUP BY 
  c.cus_qtr_key
ORDER BY 
  c.cus_qtr_key


-- Okay performance - < 15 SECS
WITH cte_results
AS
(
SELECT 
    SUBSTRING(cus_date_str, 1, 4) + '00' + datename(qq, cus_date_str) AS my_qtr_key,
    COUNT(*) as my_total
FROM 
    [FACT].[CUSTOMERS] 
GROUP BY
    SUBSTRING(cus_date_str, 1, 4) + '00' + datename(qq, cus_date_str)
)
SELECT *
FROM cte_results
ORDER BY my_qtr_key


/*

cus_qtr_key	cus_total

2011001	123032
2011002	124689
2011003	126569
2011004	126073
2012001	124540
2012002	124341
2012003	126093
2012004	124662

*/
