/******************************************************
 *
 * Name:         big-johns-bbq-table-partitioning-az.sql
 *     
 * Design Phase:
 *     Author:   John Miner
 *     Date:     04-18-2015
 *     Blog:     www.craftydba.com
 *
 *     Purpose:  Convert the Big Jon's BBQ data warehouse
 *               into a table partitioning example.
 *
 *               Implement a sliding window with hospital
 *               partitions.
 *
 ******************************************************/

 
--
-- 1 - Create primary database to house Dimension tables
--

-- Which database to use.
USE [master]
GO

-- Delete existing database
IF  EXISTS (SELECT name FROM sys.databases WHERE name = N'BBQ_TABLE_PART')
DROP DATABASE BBQ_TABLE_PART
GO

-- Create new database
CREATE DATABASE BBQ_TABLE_PART
(
MAXSIZE = 20GB,
EDITION = 'STANDARD',
SERVICE_OBJECTIVE = 'S2'
)
GO  

-- Show the new database
SELECT * FROM sys.databases WHERE name = 'BBQ_TABLE_PART';
GO


--
-- 2 - Create Fact schema
--

-- Which database to use.
USE [BBQ_TABLE_PART]
GO

-- Delete existing schema.
IF EXISTS (SELECT * FROM sys.schemas WHERE name = N'FACT')
DROP SCHEMA [FACT]
GO

-- Add new schema.
CREATE SCHEMA [FACT] AUTHORIZATION [dbo]
GO

-- Show the new schema
SELECT * FROM sys.schemas WHERE name = 'FACT';
GO


--
-- 3 - Create Dimension schema
--

-- Which database to use.
USE [BBQ_TABLE_PART]
GO

-- Delete existing schema.
IF EXISTS (SELECT * FROM sys.schemas WHERE name = N'DIM')
DROP SCHEMA [DIM]
GO

-- Add new schema.
CREATE SCHEMA [DIM] AUTHORIZATION [dbo]
GO

-- Show the new schema
SELECT * FROM sys.schemas WHERE name = 'DIM';
GO


--
-- 4 - Create date dim table 
--

-- Use the correct database
USE [BBQ_TABLE_PART];
GO

-- Delete existing table
IF  OBJECT_ID(N'[DIM].[DIM_DATE]') > 0
    DROP TABLE [DIM].[DIM_DATE]
GO

-- Create new table
CREATE TABLE [DIM].[DIM_DATE]
(
  date_key INT IDENTITY(1,1) NOT NULL CONSTRAINT PK_DIM_DATE PRIMARY KEY,
  date_date SMALLDATETIME NOT NULL,
  date_string VARCHAR(10) NOT NULL,
  date_month VARCHAR(3) NOT NULL,
  date_day VARCHAR(3) NOT NULL,
  date_int_qtr INT NOT NULL,
  date_int_doy INT NOT NULL,
  date_int_yyyy INT NOT NULL,
  date_int_mm INT NOT NULL,
  date_int_dd INT NOT NULL
)
ON [PRIMARY]
WITH (DATA_COMPRESSION = PAGE);
GO

-- Drop existing index
IF EXISTS (SELECT name FROM sys.indexes WHERE name = N'IX_DIM_DATE_STR')
DROP INDEX IX_DIM_DATE_STR ON [DIM].[DIM_DATE];
GO

-- Create new index
CREATE INDEX IX_DIM_DATE_STR
    ON [DIM].[DIM_DATE] (date_string); 
GO


--
-- 5 - Create pig packages
--

-- Use the correct database
USE [BBQ_TABLE_PART];
GO

-- Delete existing table
IF  EXISTS (
    SELECT * FROM sys.objects
    WHERE object_id = OBJECT_ID(N'[DIM].[PIG_PACKAGES]') AND
    type in (N'U'))
DROP TABLE [DIM].[PIG_PACKAGES]
GO

-- Create new table
CREATE TABLE [DIM].[PIG_PACKAGES]
(
   pkg_id int not null identity (1, 1),
   pkg_name varchar(40)
);
GO

-- Alter the table (primary key)
ALTER TABLE [DIM].[PIG_PACKAGES]
   ADD CONSTRAINT PK_PIG_PACKAGES_ID PRIMARY KEY CLUSTERED (pkg_id);
GO


--
-- Does not work in Azure
--

/*
INSERT INTO [DIM].[PIG_PACKAGES] ([pkg_name]) 
  SELECT ([pkg_name]) FROM [BIG_JONS_BBQ_DW].[DIM].[PIG_PACKAGES] WITH (NOLOCK);
*/


--
-- 6 - Load Pig Packages - Dimension Table
--

-- Import data - load-package-info.bat  
SELECT TOP 100 * FROM [DIM].[PIG_PACKAGES]
GO


--
-- 7 - Load Date - Dimension Table
--

-- Import data - load-pig-info.bat  
SELECT TOP 100 * FROM [DIM].[DIM_DATE]
GO


--
-- 8 - Create partition function & schema
-- 

-- Which database to use.
USE [BBQ_TABLE_PART]
GO

-- Create the partition function - (2011 & 2012) data
CREATE PARTITION FUNCTION PF_HASH_BY_VALUE (INT) AS RANGE LEFT 
FOR VALUES 
(
2010004,
2011001,	
2011002,	
2011003,	
2011004,	
2012001,	
2012002,	
2012003,	
2012004,
2013001
)
GO

-- Show the function
SELECT * FROM sys.partition_functions
GO

-- Create the partition scheme
CREATE PARTITION SCHEME PS_HASH_BY_VALUE 
AS PARTITION PF_HASH_BY_VALUE
ALL TO ([PRIMARY]);
GO

-- Show the scheme
SELECT * FROM sys.partition_schemes
GO


--
-- 9 - Create table using partition scheme & page compression
-- 

CREATE TABLE [FACT].[CUSTOMERS]
(
	[cus_id] [int] NOT NULL,
	[cus_lname] [varchar](40) NULL,
	[cus_fname] [varchar](40) NULL,
	[cus_phone] [char](12) NULL,
	[cus_address] [varchar](40) NULL,
	[cus_city] [varchar](20) NULL,
	[cus_state] [char](2) NULL,
	[cus_zip] [char](5) NOT NULL,
	[cus_package_key] [int] NOT NULL,
	[cus_start_date_key] [int] NOT NULL,
	[cus_end_date_key] [int] NOT NULL,
	[cus_date_str] [varchar](10) NOT NULL,
	[cus_qtr_key] int not null,

    CONSTRAINT [PK_CUSTOMERS_ID] PRIMARY KEY CLUSTERED 
    ([cus_id], [cus_qtr_key])
) 
ON PS_HASH_BY_VALUE ([cus_qtr_key])
WITH (DATA_COMPRESSION = PAGE)
GO


--
-- 10 - Load the partitioned table with data
--

-- Import data - load-pig-info.bat  
SELECT TOP 100 * FROM [FACT].[CUSTOMERS]
GO


--
-- 11 - Validate data placement
--

-- Use dmv to get partitions
SELECT 
  Partition_Number, Row_Count
  --, * 
FROM sys.dm_db_partition_stats
WHERE object_id = object_id('FACT.CUSTOMERS'); 
GO

-- Using the $PARTITION function
SELECT 
    [cus_qtr_key],
    $PARTITION.PF_HASH_BY_VALUE([cus_qtr_key]) as Partition_Number, 
    COUNT(*) as Row_Count
FROM 
    [BBQ_TABLE_PART].[FACT].[CUSTOMERS]
GROUP BY 
    [cus_qtr_key],
    $PARTITION.PF_HASH_BY_VALUE([cus_qtr_key])
ORDER BY
    [cus_qtr_key] 
GO


--
-- 12 - Create partition hash key table for sliding window
--

-- Use the correct database
USE [BBQ_TABLE_PART];
GO

-- Delete existing table
IF  EXISTS (
    SELECT * FROM sys.objects
    WHERE object_id = OBJECT_ID(N'[DIM].[PARTITIONS]') AND
    type in (N'U'))
DROP TABLE [DIM].[PARTITIONS]
GO

-- Create new table
CREATE TABLE [DIM].[PARTITIONS]
(
   part_id int not null identity (1, 1),
   part_hash_key int not null
);
GO

-- Alter the table (primary key)
ALTER TABLE [DIM].[PARTITIONS]
   ADD CONSTRAINT PK_PARTITIONS_ID PRIMARY KEY CLUSTERED (part_id);
GO

-- Alter the table (primary key)
ALTER TABLE [DIM].[PARTITIONS]
   ADD CONSTRAINT UNQ_PARTITIONS_HASH UNIQUE (part_hash_key);
GO

-- Add data to the table
INSERT INTO [DIM].[PARTITIONS] (part_hash_key)
VALUES
    (2010001),
    (2010002),
    (2010003),
    (2010004),
    (2011001),
    (2011002),
    (2011003),
    (2011004),
    (2012001),
    (2012002),
    (2012003),
    (2012004),
    (2013001),
    (2013002),
    (2013003),
    (2013004),
    (2014001),
    (2014002),
    (2014003),
    (2014004),
    (2015001),
    (2015002),
    (2015003),
    (2015004);
GO


--
-- 13 - Create a function to add a new partition
--

-- Delete existing procedure
IF OBJECT_ID(N'[dbo].[USP_ADD_CUSTOMER_PARTITION]') > 0
    DROP PROCEDURE [dbo].[USP_ADD_CUSTOMER_PARTITION]
GO

-- Create new procedure
CREATE PROCEDURE [dbo].[USP_ADD_CUSTOMER_PARTITION]
AS

    -- No info messages
    SET NOCOUNT ON;

    -- Declare variables
    DECLARE @VAR_VAL1 INT;
    DECLARE @VAR_STR1 VARCHAR(128);
    DECLARE @VAR_SQL1 VARCHAR(MAX);

    DECLARE @QTR_KEY INT;
    DECLARE @PART_KEY INT;
	
    -- Get the keys
    SELECT @QTR_KEY = MAX([cus_qtr_key]) FROM [BBQ_TABLE_PART].[FACT].[CUSTOMERS];
    SELECT @PART_KEY = [part_id] FROM [DIM].[PARTITIONS] WHERE [part_hash_key] = @QTR_KEY;

    -- Get the next hash (assumes empty hospital parts)
    SELECT 
        @VAR_VAL1 = [part_id],
        @VAR_STR1 = [part_hash_key]
    FROM 
        [DIM].[PARTITIONS] AS P1
    WHERE
	    P1.part_id =  @PART_KEY + 2;  

    -- Dynamic SQL
    SET @VAR_SQL1 = '';
    SET @VAR_SQL1 = @VAR_SQL1 + 'ALTER PARTITION FUNCTION PF_HASH_BY_VALUE () ';    
    SET @VAR_SQL1 = @VAR_SQL1 + 'SPLIT RANGE (' + LTRIM(RTRIM(@VAR_STR1)) + ');';    
    PRINT @VAR_SQL1;

	--  Make the partition
    EXEC(@VAR_SQL1);

GO

-- Test the function
-- EXEC [dbo].[USP_ADD_CUSTOMER_PARTITION]
GO

-- Must have installed helper view
SELECT 
    [partition_number]
   ,[rows]
   ,[index_name]
   ,[index_type_desc]
   ,[type_desc]
   ,[parameter_id]
   ,[value]
FROM
   [dbo].[partitioned_objects_range_values]
WHERE
   [value] IS NOT NULL
ORDER BY
   [value]
GO

-- Before hospital partition
INSERT INTO [FACT].[CUSTOMERS] VALUES
(1000000, 'Johnson', 'David', '787-453-7023', 'Abbotswell Road', 'Guayama', 'PR', '00784', 1, 2, 366, '2010-12-02',	2010004);


--
-- 13 - Create a function to add a new table
--

-- Delete existing procedure
IF OBJECT_ID(N'[dbo].[USP_ADD_CUSTOMER_TEMP_TBL]') > 0
    DROP PROCEDURE [dbo].[USP_ADD_CUSTOMER_TEMP_TBL]
GO

-- Create new procedure
CREATE PROCEDURE [dbo].[USP_ADD_CUSTOMER_TEMP_TBL]
AS

    -- No info messages
    SET NOCOUNT ON;

    -- Declare variables
    DECLARE @VAR_SQL2 VARCHAR(MAX);

    -- Create temp table on oldest partition
    SET @VAR_SQL2 = '';
    SET @VAR_SQL2 = @VAR_SQL2 + ' CREATE TABLE [FACT].[TMP_CUSTOMERS] ';
    SET @VAR_SQL2 = @VAR_SQL2 + '( ';
	SET @VAR_SQL2 = @VAR_SQL2 + '    [cus_id] [int] NOT NULL, ';
	SET @VAR_SQL2 = @VAR_SQL2 + '    [cus_lname] [varchar](40) NULL, ';
	SET @VAR_SQL2 = @VAR_SQL2 + '    [cus_fname] [varchar](40) NULL, ';
	SET @VAR_SQL2 = @VAR_SQL2 + '    [cus_phone] [char](12) NULL, ';
	SET @VAR_SQL2 = @VAR_SQL2 + '    [cus_address] [varchar](40) NULL, ';
	SET @VAR_SQL2 = @VAR_SQL2 + '    [cus_city] [varchar](20) NULL, ';
	SET @VAR_SQL2 = @VAR_SQL2 + '    [cus_state] [char](2) NULL, ';
	SET @VAR_SQL2 = @VAR_SQL2 + '    [cus_zip] [char](5) NOT NULL, ';
	SET @VAR_SQL2 = @VAR_SQL2 + '    [cus_package_key] [int] NOT NULL, ';
	SET @VAR_SQL2 = @VAR_SQL2 + '    [cus_start_date_key] [int] NOT NULL, ';
	SET @VAR_SQL2 = @VAR_SQL2 + '    [cus_end_date_key] [int] NOT NULL, ';
	SET @VAR_SQL2 = @VAR_SQL2 + '    [cus_date_str] [varchar](10) NOT NULL, ';
	SET @VAR_SQL2 = @VAR_SQL2 + '    [cus_qtr_key] int not null, ';

    SET @VAR_SQL2 = @VAR_SQL2 + '    CONSTRAINT [PK_CUSTOMERS_TEMP_ID] PRIMARY KEY CLUSTERED  ';
    SET @VAR_SQL2 = @VAR_SQL2 + '    ([cus_id], [cus_qtr_key]) ';
    SET @VAR_SQL2 = @VAR_SQL2 + ') ';

    SET @VAR_SQL2 = @VAR_SQL2 + '	ON [PRIMARY]';
    SET @VAR_SQL2 = @VAR_SQL2 + '	WITH (DATA_COMPRESSION = PAGE);';
  
    PRINT @VAR_SQL2;
    EXEC(@VAR_SQL2);

GO

-- Test the function
-- EXEC [dbo].[USP_ADD_CUSTOMER_TEMP_TBL]
GO

-- After hospital partition
INSERT INTO [FACT].[CUSTOMERS] VALUES
(1000001, 'Johnson', 'David', '787-453-7023', 'Abbotswell Road', 'Guayama', 'PR', '00784', 1, 2, 366, '2010-12-02',	2013001);


--
-- 14 - Create a function to drop a old partition
--

-- Delete existing procedure
IF OBJECT_ID(N'[dbo].[USP_DEL_CUSTOMER_PARTITION]') > 0
    DROP PROCEDURE [dbo].[USP_DEL_CUSTOMER_PARTITION]
GO

-- Create new procedure
CREATE PROCEDURE [dbo].[USP_DEL_CUSTOMER_PARTITION]
AS

  -- No info messages
  SET NOCOUNT ON;

  -- Declare variables
  DECLARE @VAR_VAL3 INT;
  DECLARE @VAR_STR3 VARCHAR(128);  

  DECLARE @VAR_VAL4 INT;
  DECLARE @VAR_STR4 VARCHAR(128); 
	  
  DECLARE @VAR_SQL5 VARCHAR(MAX);
  DECLARE @RUN BIT = 1;

  DECLARE @QTR_KEY INT;
  DECLARE @PART_KEY INT;
	
  -- Get the keys
  SELECT @QTR_KEY = MIN([cus_qtr_key]) FROM [BBQ_TABLE_PART].[FACT].[CUSTOMERS];
  SELECT @PART_KEY = [part_id] FROM [DIM].[PARTITIONS] WHERE [part_hash_key] = @QTR_KEY;

  -- Get information for partition 1
  SELECT 
    @VAR_VAL3 = [part_id],
    @VAR_STR3 = [part_hash_key]
  FROM 
    [DIM].[PARTITIONS] AS P1
  WHERE
	P1.part_id = @PART_KEY;
	  
  -- Get information for partition 2
  SELECT 
    @VAR_VAL4 = [part_id],
    @VAR_STR4 = [part_hash_key]
  FROM 
    [DIM].[PARTITIONS] AS P1
  WHERE
	P1.part_id = @PART_KEY + 1;

  -- Add new temporary table on partition 2
  SET @VAR_SQL5 = 'EXEC [dbo].[USP_ADD_CUSTOMER_TEMP_TBL] ';
  PRINT @VAR_SQL5;
  IF (@RUN=1) EXEC(@VAR_SQL5);
          
  -- Move data from part 2 to temporary table
  SET @VAR_SQL5 = 'ALTER TABLE [FACT].[CUSTOMERS] SWITCH PARTITION 2 TO [FACT].[TMP_CUSTOMERS]';
  PRINT @VAR_SQL5;
  IF (@RUN=1) EXEC(@VAR_SQL5);

  -- Remove from function - part 1
  SET @VAR_SQL5 = 'ALTER PARTITION FUNCTION PF_HASH_BY_VALUE () MERGE RANGE (' + STR(@VAR_STR3) + ');';
  PRINT @VAR_SQL5;
  IF (@RUN=1) EXEC(@VAR_SQL5);

  -- Remove from function - part 2
  SET @VAR_SQL5 = 'ALTER PARTITION FUNCTION PF_HASH_BY_VALUE () MERGE RANGE (' + STR(@VAR_STR4) + ');';
  PRINT @VAR_SQL5;
  IF (@RUN=1) EXEC(@VAR_SQL5);

  -- drop the temporary table
  SET @VAR_SQL5 = 'DROP TABLE [FACT].[TMP_CUSTOMERS];'
  PRINT @VAR_SQL5;  
  IF (@RUN=1) EXEC(@VAR_SQL5);

  -- Set hospital before as new place to store data
  SET @VAR_SQL5 = 'ALTER PARTITION SCHEME PS_HASH_BY_VALUE NEXT USED [PRIMARY];';
  PRINT @VAR_SQL5;  
  EXEC(@VAR_SQL5);

  -- Split hospital on old value
  SET @VAR_SQL5 = 'ALTER PARTITION FUNCTION PF_HASH_BY_VALUE () SPLIT RANGE (' + @VAR_STR4 + ');';
  PRINT @VAR_SQL5;  
  IF (@RUN=1) EXEC(@VAR_SQL5);  
GO

-- Test the function
-- EXEC [dbo].[USP_DEL_CUSTOMER_PARTITION]
GO
